import sys
from typing import Dict, Iterator, Union

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from jellyfin_apiclient_python.api import API


class Item:
    def __init__(self, item: Dict[str, str]):
        self.item = item

    def _raw_item(self) -> Dict[str, str]:
        return self.item

    def __getattr__(self, attr: str) -> Union[str, Self]:
        try:
            value = self.item[attr]
            if isinstance(value, dict):
                value = Item(value)
            return value
        except KeyError:
            raise AttributeError(f"Item has no attribute '{attr}'")


class ApiResponse:
    def __init__(self, value):
        self.value = value

    def __str__(self) -> str:
        return f"<ApiResponse object containing {len(self)} items>"

    def __iter__(self) -> Iterator[Item]:
        yield from [Item(i) for i in self.value["Items"]]

    def __getitem__(self, key) -> Item:
        return Item(self.value["Items"][key])

    def __len__(self) -> int:
        return self.value["TotalRecordCount"] + self.value["StartIndex"]


class WrappedAPI(API):
    def _get(self, handler, params=None):
        return ApiResponse(super()._get(handler, params=params))




import argparse
import json
from getpass import getpass as getpassword
from getpass import getuser

from .client import CREDENTIALS_FILE, auth_with_password, create_client


def create_jellyfin_token() -> None:
    client = create_client("create_jellyfin_token")
    parser = argparse.ArgumentParser()
    parser.add_argument("-u", "--user", dest="user", default=getuser())
    parser.add_argument(
        "server", nargs="?", default="https://jellyfin.wedontsleep.org/"
    )
    args = parser.parse_args()
    password = getpassword()
    result = auth_with_password(client, args.server, args.user, password)
    if "AccessToken" in result:
        credentials = client.auth.credentials.get_credentials()
        server = credentials["Servers"][0]
        server["username"] = args.user
        with open(CREDENTIALS_FILE, "w") as f:
            json.dump(server, f)
        print("Credentials saved.")



import argparse
from operator import attrgetter

from .client import authed_client
from .series import search_single_show


def unwatched() -> None:
    client = authed_client()
    parser = argparse.ArgumentParser()
    parser.add_argument("show", nargs="*")
    args = parser.parse_args()
    if not args.show:
        all_unwatched(client)
    else:
        specific_unwatched(client, " ".join(args.show))


def all_unwatched(client) -> None:
    r = client.jellyfin.search_media_items(
        term="", media="Series", limit=300
    )
    total = 0
    for series in sorted(r, key=attrgetter("Name")):
        if (count := series.UserData.UnplayedItemCount) > 0:
            ending = "s" if count != 1 else ""
            print(f"{series.Name}: {count} unwatched episode{ending}")
            total += count
    print(f"Total: {total} unwatched episodes")


def specific_unwatched(client, term: str) -> None:
    try:
        show = search_single_show(client, term)
    except ValueError as e:
        print(str(e))
        return
    unwatched = show.UserData.UnplayedItemCount
    total = sum(s.ChildCount for s in client.jellyfin.get_seasons(show.Id))
    print(f"{show.Name}: {total - unwatched} watched episodes")
    print(f"{show.Name}: {unwatched} unwatched episodes")



import argparse
from decimal import Decimal

from .client import authed_client
from .series import search_single_show


def average_duration() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("show", nargs="+")
    args = parser.parse_args()
    client = authed_client()
    try:
        show = search_single_show(client, " ".join(args.show))
    except ValueError as e:
        print(str(e))
        return
    count = 0
    duration = 0
    for season in client.jellyfin.get_seasons(show.Id):
        count += season.ChildCount
        episodes = client.jellyfin.get_season(show.Id, season.Id)
        duration += sum([e.RunTimeTicks for e in episodes])
    average = (duration / count) / int(Decimal("6E+008"))
    print(f"Average duration over {count} episodes: {average:.1f} minutes")



import unittest

from jellyash.bundle import Item


class TestItem(unittest.TestCase):
    def test_item(self):
        item = {"User": "demo"}
        wrapped = Item(item)
        self.assertDictEqual(wrapped._raw_item(), item)
        self.assertEqual(wrapped.User, "demo")
        with self.assertRaises(AttributeError):
            wrapped.user

    def test_nested_item(self):
        item = {"SessionInfo": {"UserName": "demo", "DeviceName": "bar"}}
        wrapped = Item(item)
        self.assertEqual(wrapped.SessionInfo.UserName, "demo")
        self.assertEqual(wrapped.SessionInfo.DeviceName, "bar")


class TestApiResponse(unittest.TestCase):
    pass




import json
import pathlib
import platform
import tempfile
import unittest
from unittest.mock import patch

import pytest

from jellyash import __version__
from jellyash.client import (
    auth_with_password, auth_with_token, create_client
    )


class TestClient(unittest.TestCase):
    def setUp(self):
        self.client = create_client("test")

    def test_create_client_with_app_name(self):
        data = self.client.config.data
        self.assertEqual(data['app.name'], "test")
        self.assertEqual(data['app.device_name'], platform.node())
        self.assertEqual(data['app.version'], __version__)
        self.assertTrue(data['auth.ssl'])
    
    @pytest.mark.vcr
    def test_auth_with_password(self):
        server_url = "https://demo.jellyfin.org/stable"
        auth_result = auth_with_password(self.client, server_url, "demo", "")
        self.assertEqual(auth_result["User"]["Name"], "demo")
        self.assertEqual(auth_result["SessionInfo"]["UserName"], "demo")
        self.assertEqual(
            auth_result["SessionInfo"]["DeviceName"],
            platform.node()
        )


class TestAuthWithToken(unittest.TestCase):
    def test_auth_with_token_credential_directory(self):
        with tempfile.TemporaryDirectory() as tempdir:
            ptd = pathlib.Path(tempdir)
            with patch("jellyash.client.CREDENTIALS_FILE", ptd):
                with self.assertRaises(ValueError):
                    auth_with_token(None)

    def test_auth_with_token_unreadable_file(self):
        with tempfile.NamedTemporaryFile() as tmpfile:
            ptf = pathlib.Path(tmpfile.name)
            ptf.chmod(0o0000)
            with patch("jellyash.client.CREDENTIALS_FILE", ptf):
                with self.assertRaises(PermissionError):
                    auth_with_token(None)
            ptf.chmod(0o600)

    def test_auth_with_token_invalid_json(self):
        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.file.write(b"{{\n")
            ptf = pathlib.Path(tmpfile.name)
            with patch("jellyash.client.CREDENTIALS_FILE", ptf):
                with self.assertRaises(json.decoder.JSONDecodeError):
                    auth_with_token(None)

    def test_auth_with_nonexistant_file(self):
        non_exist = pathlib.Path("/does/not/exist")
        with patch("jellyash.client.CREDENTIALS_FILE", non_exist):
            with self.assertRaises(ValueError):
                auth_with_token(None)



from .bundle import Item


def search_single_show(client, term: str) -> Item:
    r = client.jellyfin.search_media_items(term=term, media="Series")
    if len(r) == 0:
        raise ValueError(f"{term} not found")
    elif len(r) > 1:
        exact = " " not in term
        for idx, series in enumerate(r):
            if exact and term == series.Name:
                break
            elif not exact and term in series.Name:
                break
    else:
        idx = 0
    return r[idx]



import argparse

from .bundle import Item
from .client import authed_client


def episode_str(episode: Item) -> str:
    seasonindex = f"{episode.ParentIndexNumber}x{episode.IndexNumber:02}"
    return f"{episode.SeriesName} [{seasonindex}] {episode.Name}"


def nextup() -> None:
    client = authed_client()
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--limit", dest="limit", type=int, default=30)
    args = parser.parse_args()
    for episode in client.jellyfin.get_next(limit=args.limit):
        print(episode_str(episode))
